use super::{
    CommentCreated, CommentDeleted, CommentUpdated, RemoteCommentCreated, RemoteCommentUpdated,
    RemoteCommenteDeleted,
};
use crate::{store::Comment, Context, Error, Outbound, Required};
use activitystreams::{
    activity::{Announce, Create, Delete, Update},
    base::AnyBase,
    context,
    object::Note,
    prelude::*,
    public, security,
};
use std::collections::HashSet;
use url::Url;
use uuid::Uuid;

fn local_inboxes(
    submission_id: Uuid,
    comment_id: Option<Uuid>,
    ctx: &Context,
) -> Result<Vec<Url>, Error> {
    let mut urls = vec![];

    let submissioner_id = ctx
        .store
        .submissions
        .by_id(submission_id)?
        .req()?
        .profile_id();
    let submissioner_inbox = ctx
        .apub
        .endpoints_for_profile(submissioner_id)?
        .req()?
        .inbox;

    urls.push(submissioner_inbox);

    if let Some(comment_id) = comment_id {
        let commenter_id = ctx.store.comments.by_id(comment_id)?.req()?.profile_id();
        let commenter_inbox = ctx.apub.endpoints_for_profile(commenter_id)?.req()?.inbox;

        urls.push(commenter_inbox);
    }

    Ok(urls)
}

fn remote_inboxes(submissioner_id: Uuid, ctx: &Context) -> Result<Vec<Url>, Error> {
    let follower_inboxes = ctx
        .store
        .view
        .follows
        .forward_iter(submissioner_id)
        .filter_map(|follow_id| ctx.store.view.follows.right(follow_id).ok())
        .filter_map(|opt| opt)
        .filter(|follower_id| ctx.is_local(*follower_id).unwrap_or(false))
        .filter_map(|follower_id| {
            Some(
                ctx.apub
                    .endpoints_for_profile(follower_id)
                    .ok()??
                    .shared_inbox,
            )
        })
        .collect::<HashSet<_>>();

    Ok(follower_inboxes.into_iter().collect())
}

fn build_comment(
    comment: Comment,
    note_id: Url,
    person_id: Url,
    ctx: &Context,
) -> Result<AnyBase, Error> {
    let published = comment.published();
    let endpoints = ctx
        .apub
        .endpoints_for_profile(comment.profile_id())?
        .req()?;

    let mut note = Note::new();
    note.set_id(note_id)
        .set_content(comment.body())
        .set_published(published.into())
        .set_attributed_to(person_id)
        .add_to(endpoints.followers)
        .add_cc(public());
    if let Some(comment_id) = comment.comment_id() {
        let in_reply_to = ctx.apub.apub_for_comment(comment_id)?.req()?;
        note.set_in_reply_to(in_reply_to);
    }

    let any_base = note.into_any_base()?;
    ctx.apub.store_object(&any_base)?;

    Ok(any_base)
}

impl Outbound for CommentCreated {
    fn id(&self) -> Option<Uuid> {
        Some(self.comment_id)
    }

    fn inboxes(&self, ctx: &Context) -> Result<Vec<Url>, Error> {
        let comment = ctx.store.comments.by_id(self.comment_id)?.req()?;

        local_inboxes(comment.submission_id(), comment.comment_id(), ctx)
    }

    fn to_apub(&self, ctx: &Context) -> Result<AnyBase, Error> {
        let comment = ctx.store.comments.by_id(self.comment_id)?.req()?;
        let person_id = ctx.apub.apub_for_profile(comment.profile_id())?.req()?;
        let note_id = ctx.apub.info.gen_id().req()?;
        ctx.apub.comment(&note_id, self.comment_id)?;
        let note = build_comment(comment, note_id, person_id.clone(), ctx)?;

        let mut create = Create::new(person_id, note);
        create
            .set_id(ctx.apub.info.gen_id().req()?)
            .add_context(context())
            .add_context(security());
        let any_base = create.into_any_base()?;
        ctx.apub.store_object(&any_base)?;

        Ok(any_base)
    }
}

impl Outbound for RemoteCommentCreated {
    fn id(&self) -> Option<Uuid> {
        None
    }

    fn inboxes(&self, ctx: &Context) -> Result<Vec<Url>, Error> {
        remote_inboxes(self.profile_id, ctx)
    }

    fn to_apub(&self, ctx: &Context) -> Result<AnyBase, Error> {
        let comment = ctx.apub.object(&self.note_apub_id)?.req()?;
        let person_id = ctx.apub.apub_for_profile(self.profile_id)?.req()?;

        let mut announce = Announce::new(person_id, comment);
        announce
            .set_id(ctx.apub.info.gen_id().req()?)
            .add_context(context())
            .add_context(security());
        let announce = announce.into_any_base()?;
        ctx.apub.store_object(&announce)?;

        Ok(announce)
    }
}

impl Outbound for CommentUpdated {
    fn id(&self) -> Option<Uuid> {
        Some(self.comment_id)
    }

    fn inboxes(&self, ctx: &Context) -> Result<Vec<Url>, Error> {
        let comment = ctx.store.comments.by_id(self.comment_id)?.req()?;

        local_inboxes(comment.submission_id(), comment.comment_id(), ctx)
    }

    fn to_apub(&self, ctx: &Context) -> Result<AnyBase, Error> {
        let comment = ctx.store.comments.by_id(self.comment_id)?.req()?;
        let person_id = ctx.apub.apub_for_profile(comment.profile_id())?.req()?;
        let note_id = ctx.apub.apub_for_comment(comment.id())?.req()?;
        let note = build_comment(comment, note_id, person_id.clone(), ctx)?;

        let mut update = Update::new(person_id, note);
        update
            .set_id(ctx.apub.info.gen_id().req()?)
            .add_context(context())
            .add_context(security());
        let any_base = update.into_any_base()?;
        ctx.apub.store_object(&any_base)?;

        Ok(any_base)
    }
}

impl Outbound for RemoteCommentUpdated {
    fn id(&self) -> Option<Uuid> {
        None
    }

    fn inboxes(&self, ctx: &Context) -> Result<Vec<Url>, Error> {
        remote_inboxes(self.profile_id, ctx)
    }

    fn to_apub(&self, ctx: &Context) -> Result<AnyBase, Error> {
        let person_id = ctx.apub.apub_for_profile(self.profile_id)?.req()?;
        let update = ctx.apub.object(&self.update_apub_id)?.req()?;

        let mut announce = Announce::new(person_id, update);
        announce
            .set_id(ctx.apub.info.gen_id().req()?)
            .add_context(context())
            .add_context(security());
        let announce = announce.into_any_base()?;
        ctx.apub.store_object(&announce)?;

        Ok(announce)
    }
}

impl Outbound for CommentDeleted {
    fn id(&self) -> Option<Uuid> {
        None
    }

    fn inboxes(&self, ctx: &Context) -> Result<Vec<Url>, Error> {
        local_inboxes(self.submission_id, self.reply_to_id, ctx)
    }

    fn to_apub(&self, ctx: &Context) -> Result<AnyBase, Error> {
        let person_id = ctx.apub.apub_for_profile(self.profile_id)?.req()?;
        let note = ctx.apub.object(&self.note_apub_id)?.req()?;

        let mut delete = Delete::new(person_id, note);
        delete
            .set_id(ctx.apub.info.gen_id().req()?)
            .add_context(context())
            .add_context(security());
        let any_base = delete.into_any_base()?;
        ctx.apub.store_object(&any_base)?;

        Ok(any_base)
    }
}

impl Outbound for RemoteCommenteDeleted {
    fn id(&self) -> Option<Uuid> {
        None
    }

    fn inboxes(&self, ctx: &Context) -> Result<Vec<Url>, Error> {
        remote_inboxes(self.profile_id, ctx)
    }

    fn to_apub(&self, ctx: &Context) -> Result<AnyBase, Error> {
        let person_id = ctx.apub.apub_for_profile(self.profile_id)?.req()?;
        let delete = ctx.apub.object(&self.delete_apub_id)?.req()?;

        let mut announce = Announce::new(person_id, delete);
        announce
            .set_id(ctx.apub.info.gen_id().req()?)
            .add_context(context())
            .add_context(security());
        let announce = announce.into_any_base()?;
        ctx.apub.store_object(&announce)?;

        Ok(announce)
    }
}
