use actix_session::{Session, UserSession};
use actix_web::{
    dev::{Payload, Service, ServiceRequest, ServiceResponse, Transform},
    FromRequest, HttpRequest, HttpResponse, Responder,
};
use futures_core::future::LocalBoxFuture;
use futures_util::future::{ok, Ready};
use hyaenidae_accounts::AcceptedCookies;
use std::fmt;

#[derive(Clone, Debug, serde::Deserialize, serde::Serialize)]
pub struct Back {
    path: String,
}

pub(crate) struct BackPage;
pub(crate) struct BackMiddleware<S>(S);

impl FromRequest for Back {
    type Config = ();
    type Error = actix_web::Error;
    type Future = LocalBoxFuture<'static, Result<Self, Self::Error>>;

    fn from_request(req: &HttpRequest, _: &mut Payload) -> Self::Future {
        let session = req.get_session();

        Box::pin(async move {
            Ok(Back::data(&session)?.unwrap_or(Back {
                path: "/".to_owned(),
            }))
        })
    }
}

impl fmt::Display for Back {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        fmt::Display::fmt(&self.path, f)
    }
}

impl Responder for Back {
    type Error = actix_web::Error;
    type Future = Ready<Result<HttpResponse, Self::Error>>;

    fn respond_to(self, _: &HttpRequest) -> Self::Future {
        ok(HttpResponse::SeeOther()
            .header("Location", self.path)
            .finish())
    }
}

impl Back {
    fn set_data(path: String, session: &Session) -> Result<(), actix_web::Error> {
        session.set("back-data", Back { path })?;
        Ok(())
    }

    fn data(session: &Session) -> Result<Option<Self>, actix_web::Error> {
        Ok(session.get("back-data")?)
    }

    pub(crate) fn as_str(&self) -> &str {
        &self.path
    }
}

impl AsRef<str> for Back {
    fn as_ref(&self) -> &str {
        self.as_str()
    }
}

impl<S, B> Transform<S> for BackPage
where
    S: Service<Request = ServiceRequest, Response = ServiceResponse<B>, Error = actix_web::Error>,
    S::Future: 'static,
{
    type Request = S::Request;
    type Response = S::Response;
    type Error = S::Error;
    type InitError = ();
    type Transform = BackMiddleware<S>;
    type Future = Ready<Result<Self::Transform, Self::InitError>>;

    fn new_transform(&self, service: S) -> Self::Future {
        ok(BackMiddleware(service))
    }
}

impl<S, B> Service for BackMiddleware<S>
where
    S: Service<Request = ServiceRequest, Response = ServiceResponse<B>, Error = actix_web::Error>,
    S::Future: 'static,
{
    type Request = S::Request;
    type Response = S::Response;
    type Error = S::Error;
    type Future = LocalBoxFuture<'static, Result<Self::Response, Self::Error>>;

    fn poll_ready(
        &mut self,
        cx: &mut std::task::Context<'_>,
    ) -> std::task::Poll<Result<(), Self::Error>> {
        self.0.poll_ready(cx)
    }

    fn call(&mut self, req: S::Request) -> Self::Future {
        let session = req.get_session();

        let path = req
            .uri()
            .path_and_query()
            .map(|paq| paq.as_str())
            .unwrap_or("/")
            .to_owned();

        let (req, pl) = req.into_parts();
        let cookie_fut = AcceptedCookies::extract(&req);
        let req = ServiceRequest::from_parts(req, pl)
            .map_err(|_| ())
            .expect("Request has been cloned");

        let fut = self.0.call(req);

        Box::pin(async move {
            if let Err(_) = cookie_fut.await {
                return fut.await;
            }

            match fut.await {
                Ok(response) => {
                    if let Some(content_type) = response.headers().get("Content-Type") {
                        if let Ok(s) = content_type.to_str() {
                            if s == "text/html" {
                                Back::set_data(path, &session)?;
                            }
                        }
                    }
                    Ok(response)
                }
                Err(e) => Err(e),
            }
        })
    }
}
